#include "cudaErrors.cu"

#define dloc(z,x) (z)+(x)*n1gpu
#define min2(v1,v2) (((v1)<(v2))?(v1):(v2)) 

__constant__ int nt_gpu;
__constant__ float dt_gpu;
__constant__ int n1gpu;
__constant__ int n2gpu;
__constant__ int nrec_gpu;
texture <float, 2, cudaReadModeElementType> vel_ref;
cudaChannelFormatDesc cf;
cudaArray *vel_array=0;
__device__ int min3(int v1,int v2,int v3){return min2(min2(v1,v2),v3);}
__device__ int min4(int v1,int v2,int v3, int v4){return min2(min2(v1,v2),min2(v3,v4));}

float *p0_gpu, *p1_gpu, *p2_gpu, *p3_gpu, *source_gpu, *rf_gpu, *data_gpu;

#include "wave_kernel2d8o.cu"

__global__ void inject_kernel(int z, int it, float *source_gpu, float *p)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=n2gpu) return;
    p[dloc(z,i)] += source_gpu[it+i*nt_gpu];
}


__global__ void inject2_kernel(float *p1, float *rf_gpu, float *p3)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    register int j=blockIdx.y*blockDim.y+threadIdx.y;
    p3[dloc(i,j)] += p1[dloc(i,j)]*rf_gpu[dloc(i,j)];
}


__global__ void collect_kernel(int z, int ox, int it, float *data_gpu, float *p)
{
    register int i=blockIdx.x*blockDim.x+threadIdx.x;
    if(i>=nrec_gpu) return;
    data_gpu[it+(i)*nt_gpu] = p[dloc(z,i+ox)];
}


extern "C" void gpu_init(int *nt, float *dt, int *n1, int *n2, float *d1, float *d2, int *nrec, int *gpu_id)
{

    cudaSetDevice(*gpu_id);

    cuda_call(cudaMalloc(    (void**) &p0_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p1_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p2_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &p3_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &source_gpu,  (*nt)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &rf_gpu,  (*n1)*(*n2)*sizeof(float)));
    cuda_call(cudaMalloc(    (void**) &data_gpu,  (*nt)*(*nrec)*sizeof(float)));

    //bind texture reference to array
    cf = cudaCreateChannelDesc<float>();
    cuda_call(cudaMallocArray(&vel_array, &cf, *n1, *n2));

    cuda_call(cudaMemcpyToSymbol(n1gpu, n1, sizeof(int)));
    cuda_call(cudaMemcpyToSymbol(n2gpu, n2, sizeof(int)));

    cuda_call(cudaMemcpyToSymbol(nt_gpu,  nt,   sizeof(int)));

    cuda_call(cudaMemcpyToSymbol(nrec_gpu,  nrec,   sizeof(int)));

    float coeffs_cpu[COEFFS_SIZE]=get_coeffs(*d1,*d2);
    cuda_call(cudaMemcpyToSymbol(coeffs, coeffs_cpu, COEFFS_SIZE*sizeof(float)));

}

extern "C" void gpu_vel_update(float *cpu_space)
{
    int n1, n2;

    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int))); 
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int))); 
    cuda_call(cudaMemcpyToArray(vel_array, 0,0, cpu_space, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaBindTextureToArray(vel_ref, vel_array));
}

extern "C" void gpu_model(int *sou_z, float *source_cpu, int *rec_z, int *rec_ox, float *data)
{
    int nt, n1, n2, nrec;
    float *ptemp;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nrec, nrec_gpu, sizeof(int)));

    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(data_gpu, 0,nt*nrec*sizeof(float)));

    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);

    /* 11 ACTUAL WAVE PROPAGATION */

    for (int it=0; it<nt; it++)
    {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
        ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
        kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));
        kernel_exec(collect_kernel<<<nrec/80+1,80>>>(*rec_z, *rec_ox, it, data_gpu, p1_gpu));
    }

    cuda_call(cudaMemcpy(data, data_gpu, nt*nrec*sizeof(float), cudaMemcpyDeviceToHost));

}

extern "C" void gpu_model2(int *sou_z, float *source_cpu, float *rf_cpu, int *rec_z, int *rec_ox, float *data)
{
    int nt, n1, n2, nrec;
    float *ptemp;

    cuda_call(cudaMemcpyFromSymbol(&nt, nt_gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n1, n1gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&n2, n2gpu, sizeof(int)));
    cuda_call(cudaMemcpyFromSymbol(&nrec, nrec_gpu, sizeof(int)));

    cuda_call(cudaMemcpy(source_gpu, source_cpu, nt*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemcpy(rf_gpu, rf_cpu, n1*n2*sizeof(float), cudaMemcpyHostToDevice));
    cuda_call(cudaMemset(p0_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p1_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p2_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(p3_gpu, 0,n1*n2*sizeof(float)));
    cuda_call(cudaMemset(data_gpu, 0,nt*nrec*sizeof(float)));

    int nblocksx=(n1-2*FAT)/BLOCK_SIZE;
    int nblocksy=(n2-2*FAT)/BLOCK_SIZE;
    dim3 dimGrid(nblocksx,nblocksy);
    dim3 dimBlock(BLOCK_SIZE, BLOCK_SIZE);

    /* 11 ACTUAL WAVE PROPAGATION */

    for (int it=0; it<nt; it++)
    {
        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p0_gpu, p1_gpu, p0_gpu));
        ptemp=p0_gpu; p0_gpu=p1_gpu; p1_gpu=ptemp;
        kernel_exec(inject_kernel<<<n2/80+1,80>>>(*sou_z, it, source_gpu, p1_gpu));

        kernel_exec(wave_kernel<<<dimGrid, dimBlock>>>(p2_gpu, p3_gpu, p2_gpu));
        ptemp=p2_gpu; p2_gpu=p3_gpu; p3_gpu=ptemp;
        kernel_exec(inject2_kernel<<<dimGrid, dimBlock>>>(p1_gpu, rf_gpu, p3_gpu));

        kernel_exec(collect_kernel<<<nrec/80+1,80>>>(*rec_z, *rec_ox, it, data_gpu, p3_gpu));
    }

    cuda_call(cudaMemcpy(data, data_gpu, nt*nrec*sizeof(float), cudaMemcpyDeviceToHost));

}
